import datetime

__all__ = ['Field',
           'CharField',
           'IntField',
           'ArrayField',
           'BoolField',
           'DateTimeField'
           ]


class Field:
    def __init__(self, field=None, lookup=None):
        self._field = field
        self._lookup = lookup

    def __set__(self, instance, value):
        instance.__dict__['_value'] = value

    def __get__(self, instance, owner):
        value = instance.__dict__.get('_value')
        query_key = self._field if self._lookup is None else '__'.join((self._field, self._lookup))
        return {query_key: value}

    def __set_name__(self, owner, name):
        """如果field=None 设置为name"""
        if self._field is None:
            self._field = name


class CharField(Field):
    def __set__(self, instance, value):
        try:
            instance.__dict__['_value'] = str(value)
        except Exception as err:
            raise ValueError('查询参数转化为字符串失败!')


class IntField(Field):
    def __set__(self, instance, value):
        try:
            instance.__dict__['_value'] = int(value)
        except Exception as err:
            raise ValueError('值: {} 转化为整型失败!'.format(value))


class ArrayField(Field):
    def __set__(self, instance, value):
        if isinstance(value, list):
            _value = value
        elif isinstance(value, str):
            _value = value.split(',')
        elif value is None:
            _value = []
        else:
            raise ValueError('值: {} 要求是一个数组，或者逗号分割的字符串'.format(str(value)))
        instance.__dict__['_value'] = _value


class BoolField(Field):
    def __set__(self, instance, value):
        false_li, true_li = [False, 'false', '0', 0], [True, 'true', '1', 1]
        value = str(value).lower() if isinstance(value, str) else value
        if value not in false_li + true_li:
            raise ValueError('值: {} 必须为 true,"1",1 或者 false,"0",0'.format(str(value)))
        instance.__dict__['_value'] = value in true_li


class DateTimeField(Field):
    """本函数支持了使用不完整的日期时间来查询DateTimeField"""

    def __init__(self, field=None):
        self._field = field
        self._lookup = None
        super().__init__(field, None)

    def __set__(self, instance, value):
        value = value.strip()
        date, *time = value.split(' ')
        # 如果仅仅提供了日期
        if not time:
            self._handle_only_time(instance, value)
        else:  # 如果提供了时期时间，时间类型有三种1.只提供了小时 2.只提供了分钟3.完整的时分秒
            time = time[0]
            time_type = len(time.split(':'))
            # 根据time_type获取处理函数
            handle_func = getattr(self, f'_handle_type_{time_type}', self._not_support)
            handle_func(instance, value)  # 执行处理函数

    def _handle_only_time(self, instance, value):
        """如果仅仅提供了日期"""
        date_time = datetime.datetime.strptime(value, '%Y-%m-%d')
        timedelta = datetime.timedelta(days=1)
        instance.__dict__['_value'] = [date_time, date_time + timedelta]
        self._lookup = 'range'

    def _handle_type_1(self, instance, value):
        """1.只提供了小时"""
        date_time = datetime.datetime.strptime(value, '%Y-%m-%d %H')
        timedelta = datetime.timedelta(hours=1)
        instance.__dict__['_value'] = [date_time, date_time + timedelta]
        self._lookup = 'range'

    def _handle_type_2(self, instance, value):
        """2.只提供了分钟"""
        date_time = datetime.datetime.strptime(value, '%Y-%m-%d %H:%M')
        timedelta = datetime.timedelta(minutes=1)
        instance.__dict__['_value'] = [date_time, date_time + timedelta]
        self._lookup = 'range'

    def _handle_type_3(self, instance, value):
        """3.完整的时分秒"""
        date_time = datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
        instance.__dict__['_value'] = date_time
        self._lookup = None

    def _not_support(self, instance, value):
        raise ValueError('{} 必须为合法的日期时间格式，请使用 YYYY-MM-DD 或 '
                         'YYYY-MM-DD HH:MM:SS'.format(value))



from django.conf import settings

__all__ = ['Field',
           'FileField',
           'ImageField',
           'DateTimeField',
           'JSONField',
           'Foreignkey',
           'ManyToManyField',
           'RelatedField',
           'AttrField']


class Field:
    def __init__(self):
        self.field = None

    def __get__(self, instance, owner):
        return self.process

    def __set_name__(self, owner, name):
        if self.field is None:
            self.field = name

    @classmethod
    def process(cls, value):
        # 整型和None类型不转换，布尔类型也是整型
        if isinstance(value, int) or value is None:
            return value
        else:
            return str(value)


class FileField(Field):

    def process(self, value):
        if not value:
            return None

        return getattr(settings, 'MEDIA_URL', '') + str(value)


class ImageField(FileField):
    """直接继承FileField"""


class DateTimeField(Field):

    def process(self, value):
        if not value:
            return value
        return str(value)[0:19]


class JSONField(Field):
    def process(self, value):
        if not value:
            return value
        return value


class AttrField(Field):
    """有时候需要的数据并非都定义在model的fields中，比如注解信息，这时候就需要访问模型对象的属性获得"""

    def __init__(self, attr_name=None):
        super().__init__()
        self.field = attr_name
        
    def process(self, model_obj):
        return getattr(model_obj, self.field, None)


class Foreignkey(Field):
    def __init__(self, serialize_class):
        super().__init__()
        self.serialize_class = serialize_class

    def process(self, value):
        if not value:
            return None
        return self.serialize_class(value)


class ManyToManyField(Field):
    def __init__(self, serialize_class):
        super().__init__()
        self.serialize_class = serialize_class

    def __get__(self, instance, owner):
        self.instance = instance
        return self.process

    def process(self, model_obj):
        # 钩子函数名
        hook_name = 'hook_' + self.field
        hook_func = getattr(self.instance, hook_name, None)
        if hook_func:
            query_set = hook_func(model_obj)
        else:
            many_obj = getattr(model_obj, self.field, None)
            query_set = many_obj.all() if many_obj else ()
        return self.serialize_class(query_set)


class RelatedField(Field):
    """" 序列化相关表
    class A(models.Model):
        name = models.CharField(max_length=50)

    class B(models.Model):
        a = models.ForeignKey(A)

    在上述示例中model A作为外键被model B所关联，那么在序列化model A时可以使用RelatedField来添加model B的序列化:
    class ASerializers(Serializers):
        # b 是model B的类名的小写
        b = RelatedField(serialize_class=xxx)
        # 也可以使用"别名b", 但是需要传入related_name参数 一般为类名小写加_set本例中b_set
        别名b = RelatedField(serialize_class=xxx, related_name="b_set")

    当关联字段在model定义了related_name时：
    class B(models.Model):
        a = models.ForeignKey(A,related_name='related_b')

    # related_name需要和model中定义的related_name一致
    class ASerializers(Serializers):
        b = RelatedField(serialize_class=xxx, related_name="related_b")
    """

    def __init__(self, serialize_class, related_name=None):
        super().__init__()
        self.related_name = related_name
        self.serialize_class = serialize_class

    def __get__(self, instance, owner):
        self.instance = instance
        return self.process

    def __set_name__(self, owner, name):
        if self.related_name is None:
            self.related_name = name + '_set'
        self.name = name  # 用于找到钩子函数名

    def process(self, model_obj):
        # 钩子函数名
        hook_name = 'hook_' + self.name
        hook_func = getattr(self.instance, hook_name, None)
        if hook_func:
            query_set = hook_func(model_obj)
        else:
            cross_obj = getattr(model_obj, self.related_name, None)
            query_set = cross_obj.all() if cross_obj else ()
        return self.serialize_class(query_set)



import types
from django.http import HttpRequest
from functools import partial

__all__ = ['form_validator',
           'RequestMethodError',
           'FormValidationError']


class RequestMethodError(Exception):
    """请求方法错误"""


class FormValidationError(Exception):
    """表单验证失败"""


def form_validator(form=None, *, must_method: str = None):
    """
    将django的原生forms.Form包装成一个表单验证装饰器，下面是演示：

    # 1.Django Form 变成了一个表单验证装饰器
    @form_validator
    class LoginForm(forms.Form):
        username = forms.CharField(required=True)
        password = forms.CharField(required=True)

    # 2.在接口函数中使用AppForm
    @LoginForm
    def login_view(request):
        pass

    # 3.指定请求方法
    @form_validator(must_method)
    class LoginForm(forms.Form):
        pas

    # 4.定义Form时从其它Form中摘取字段
    @form_validator
    class A(forms.Form): # 定义一个AForm
        username = forms.CharField(required=True)

        def clean_username(self):
            return self.cleaned_data['username']


    @form_validator
    class B(forms.Form): # BForm从AForm中摘取username和clean_username
        username = A.username
        clean_username = A.clean_username
    """
    if must_method:
        must_method = must_method.upper()
    if form:
        return FormValidator(form, must_method=must_method)
    else:
        return partial(FormValidator, must_method=must_method)


class FormValidator:
    def __init__(self, form=None, *, must_method: str = None):
        self.Form = form
        self.must_method = must_method

    def __call__(self, _callable):
        # 判断被装饰的是普通函数还是APIView
        is_func = True if isinstance(_callable, types.FunctionType) else False
        func = _callable if is_func else _callable.api

        def wrapped(*args, **kwargs):
            request = next(i for i in args if isinstance(i, HttpRequest))
            # 检查请求方法
            if self.must_method and request.method != self.must_method:
                raise RequestMethodError(f'必须为{self.must_method}请求！')
            form = self.Form(getattr(request, request.method))
            if not form.is_valid():  # 表单验证不通过直接抛出异常
                raise FormValidationError(dict(form.errors))
            request.cleaned_data = form.cleaned_data  # 添加cleaned_data
            return func(*args, **kwargs)

        if not is_func:  # 如果不是函数则认为是APIView
            _callable.api = wrapped
        return wrapped if is_func else _callable

    def __getattr__(self, item):
        if item.startswith('clean_'):
            return getattr(self.Form, item)
        try:
            return self.Form.base_fields[item]
        except KeyError:
            raise AttributeError(f"'{self.Form.__name__}' has no attribute '{item}'")

# class FormBase(type):
#     def __new__(mcs, name, bases, attrs):
#         # 所有metaclass=FormBase的类，都是FormBase的实例（元类的实例是类）
#         new_class = super().__new__(mcs, name, bases, attrs)
#         parents = [b for b in bases if isinstance(b, FormBase)]
#         if not parents:  # 排除模型类本身
#             return new_class
#
#         fields = {}
#         for key, attr in attrs.items():
#             if isinstance(attr, forms.Field) or callable(attr):
#                 fields[key] = attr
#
#         for base in reversed(parents):
#             base_fields = getattr(base, '_fields', dict())
#             for key, value in base_fields.items():
#                 if key not in fields:
#                     fields[key] = value
#
#         new_class._fields = fields
#         new_class.Form = type('Form', (forms.Form,), copy.deepcopy(fields))
#         return new_class
#
#     def __call__(cls, *args, **kwargs):
#         instance = super().__call__(*args, **kwargs)
#         if instance._is_func:  # 函数直接实例，实例是callable的
#             return instance
#         else:  # 非函数即认为是APIView
#             instance._callable.api = instance
#             return instance._callable
#
#
# class FormValidator(metaclass=FormBase):
#     # 指定的请求方法，为None则不限制请求方法
#     must_method = None  # POST, GET, PUT...
#
#     def __init__(self, _callable):
#         # 判断被装饰的是普通函数还是APIView
#         self._is_func = True if isinstance(_callable, types.FunctionType) else False
#         self._callable = _callable
#         self._func = _callable if self._is_func else _callable.api
#         # 唯一的作用是为了消灭编辑器在编写clean_xxx函数时调用self.cleaned_data时出现的高亮提示
#         self.cleaned_data = {}
#         self.data = {}
#
#     def __call__(self, *args, **kwargs):
#         request = next(i for i in args if isinstance(i, HttpRequest))
#         # 检查请求方法
#         if self.must_method and request.method != self.must_method:
#             raise RequestMethodError(f'必须为{self.must_method}请求！')
#         # 检查表单验证
#         form = self.Form(getattr(request, request.method))
#         if not form.is_valid():
#             raise FormValidationError(dict(form.errors))
#         # 添加cleaned_data到request中
#         request.cleaned_data = form.cleaned_data
#         return self._func(*args, **kwargs)
#
#     def __get__(self, instance, cls):
#         if instance is None:
#             return self
#         else:
#             return types.MethodType(self, instance)
#
#     class Meta:
#         model = Orders._meta.fields
#         fields = ['pub_date', 'headline', 'content', 'reporter']



from typing import *
from django.http import JsonResponse
from django.db.models import QuerySet

__all__ = ["APIView"]


class HookClassMixin:
    """钩子函数类"""

    def process_request(self, request):
        """拦截处理request"""
        return request

    def inject_queryset(self, request):
        """ 注入queryset
        inject_queryset需要返回一个queryset,
        该函数在过滤操作之前运行，如果返回了一个queryset，过滤函数会从这个queryset中过滤出结果，
        如果没有定义该函数，或者该函数返回None，queryset就是过滤类中的Meta.model.objects.all()
        可以在该函数中引发异常，使接口返回400错误，表示资源不可访问或不存在。
        注意：
            1.inject_queryset不能对queryset进行切片等破坏性操作，切片会禁止对queryset进行进一步的排序或过滤！
            2.当inject_queryset未定义或返回None时, filter_class一定不能为None！
        """
        return None

    def process_queryset(self, queryset):
        """对过滤之后，分页之前的queryset进行处理"""
        return queryset

    def process_results(self, results):
        """results的拦截函数，可以重写该函数对results 进行修改
        results是序列化后的结果，results一般是由多个字典对象组成的列表或者一个单独的字典"""
        return results

    def process_data(self, data):
        """data拦截函数，data是接口返回给请求的最终数据"""
        return data


class APIView(HookClassMixin):
    # filter_class为None时，一定要定义process_queryset函数，且process_queryset不能返回None
    filter_class = None
    paginator_class = None
    serializer_class = None

    def __init__(self):
        self.request = None
        self.queryset = None

    def _filtering(self, user_query: dict, queryset):
        """过滤，从request中获取查询条件(query),
        如果inject_queryset返回的queryset为None,queryset就是filter_class中的Meta.model.objects.all()"""
        if self.filter_class:
            return self.filter_class(user_query, queryset)
        else:
            return queryset

    def _paging(self, queryset, request) -> Tuple[QuerySet, Dict or None]:
        if self.paginator_class is None:
            return queryset, None
        paginator = self.paginator_class(queryset, request)
        return paginator.sub_set(), paginator.page_count()

    def _build_data(self, results: Dict or List, page_count: Dict or None):
        """生成最终数据，可以被序列化的数据"""
        if page_count:
            data = {'results': results}
            data.update(page_count)
        else:
            data = results
        return data

    def api(self, request):
        """直接暴露给路由的API接口"""
        request = self.process_request(request)  # 处理request
        self.request = request
        # 注入queryset
        try:
            self.queryset = self.inject_queryset(request)
        except Exception as err:
            return JsonResponse({'msg': '资源不可访问，或不存在！', 'errors': str(err)}, status=400)

        if self.queryset is None and self.filter_class is None:
            raise RuntimeError(f'{self.__class__.__name__} inject_queryset返回None,且没有定义filter_class')

        # 过滤
        user_query = getattr(request, request.method)
        try:
            self.queryset = self._filtering(user_query, self.queryset)
        except Exception as err:
            return JsonResponse({'msg': 'filter error！', 'errors': str(err)}, status=400)

        # 处理queryset
        try:
            self.queryset = self.process_queryset(self.queryset)
        except Exception as err:
            return JsonResponse({'msg': 'process queryset error!', 'errors': str(err)}, status=400)

        # 翻页
        # page_count: Dict or None  # page_count为None则证明无需翻页(paginator_class is None)，直接返回queryset
        self.queryset = self.queryset
        self.queryset, page_count = self._paging(self.queryset, request)
        # 序列化
        try:
            results = self.serializer_class(self.queryset)
        except Exception as err:
            return JsonResponse({'msg': 'serializer error！', 'errors': str(err)}, status=400)

        # 处理results
        results = self.process_results(results)

        # 生成data（self.results中填入分页数据）
        data = self._build_data(results, page_count)

        # 处理data
        data = self.process_data(data)
        return JsonResponse(data, safe=False)

    @classmethod
    def as_view(cls):
        def api(request):
            self = cls()
            return self.api(request)

        return api

    def __init_subclass__(cls, **kwargs):
        if not cls.serializer_class:
            raise TypeError(f'{cls.__name__}没有配置serializer_class！')



from .fields import *
import importlib
from django.db.models import Model
from collections import Iterable

__all__ = ['Serializer']


class SerializerMeta(type):
    def __new__(mcs, name, bases, attrs: dict):
        # 所有metaclass=SerializerMeta的类，都是SerializerMeta的实例（元类的实例是类）
        parents = [b for b in bases if isinstance(b, SerializerMeta)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        _meta = mcs.build_meta(name, parents, attrs)
        new_class = super().__new__(mcs, name, bases, attrs)
        new_class._meta = _meta

        # 普通字段描述符在这里自动添加，外键，多对多，关联表等特殊字段在子类中定义
        this_module = importlib.import_module(__name__)  # this_module 自身模块
        model: Model = _meta.model  # 取出meta中的model
        for field in model._meta.fields:  # 定义描述符
            # 忽略ForeignKey,如果需要嵌套序列化外键必须在子类中定义
            if field.__class__.__name__ == 'ForeignKey':
                continue
            # 等价于:descriptor =  CharField DateTimeField，如果没有则取Field
            descriptor = getattr(this_module, field.__class__.__name__, Field)
            # 等价于 name = CharField()
            setattr(new_class, field.name, descriptor())

        # model._meta.fields以外的字段，多对多，关联表，和注解等无法通过遍历模型的_meta.fields获取
        out_fields = set()
        for attr_name, attr in attrs.items():
            if isinstance(attr, (ManyToManyField, RelatedField, AttrField)):
                out_fields.add(attr_name)
        # 处理特多对多和关联表的继承
        for p in parents:
            p_out_fields = getattr(p, '_out_fields', set())
            for i in p_out_fields:
                out_fields.add(i)
        new_class._out_fields = out_fields

        return new_class

    @classmethod
    def build_meta(mcs, name, parents: list, attrs: dict):
        """
        生成类的_meta信息
        :param name:类名称
        :param parents:继承的父类列表
        :param attrs:属性字典
        :return:meta class
        """
        # 如果自身没有Meta Class 就从父组件取，当然直接继承自Serializer自己也没有定义Meta class时也是徒劳为None，
        # 但是这种情况并不会发生，直接继承Serializer一定会定义Meta class 因为Model是必填的
        own_meta = attrs.pop('Meta', None)
        # 直接继承Serializer这里parent_meta为None,但是上一行不可能为None
        parent_meta = getattr(parents[-1], '_meta', None)
        if own_meta:
            # 如果自身没有在Meta中定义model，且父类也没有定义，会抛出异常
            if not getattr(own_meta, 'model', None) and not getattr(parent_meta, 'model', None):
                raise AttributeError(f'序列化类:{name} 没有在Class Meta中指定model')

            if not getattr(own_meta, 'model', None):  # 如果自身没有在Meta中定义model则从父类中继承（右侧继承）
                own_meta.model = parent_meta.model
        meta = own_meta or parent_meta
        # 深度复制meta, 防止子类继承父类的meta时，子类对meta进行修改影响到父类
        meta = type('_meta', (), dict(meta.__dict__))
        meta = mcs.meta_set_defaults(meta)
        return meta

    @classmethod
    def meta_set_defaults(mcs, meta):
        null = object()  # 空值
        meta_defaults = dict(only=(), exclude=())
        # 当子组件没有设置meta的部分属性时给一个默认值
        for attr_name, value in meta_defaults.items():
            if getattr(meta, attr_name, null) is null:
                setattr(meta, attr_name, value)
        return meta

    def __call__(cls, set_or_obj):
        self = super().__call__()
        return self._serialize(set_or_obj)


class Serializer(metaclass=SerializerMeta):

    def _serialize(self, set_or_obj):
        """序列化queryset或model_obj"""

        if isinstance(set_or_obj, Iterable):
            return [self._to_dict(obj) for obj in set_or_obj]

        elif isinstance(set_or_obj, Model):
            return self._to_dict(set_or_obj)

        else:
            raise TypeError('{}类型不支持序列化'.format(type(set_or_obj)))

    def _to_dict(self, model_obj):
        data_dict = {}

        for field in model_obj._meta.fields:
            # only被定义时，忽略only以外的全部字段
            if self._meta.only and field.name not in self._meta.only:
                continue
            # 只在没定义only时defer生效，only一旦定义defer失效
            if not self._meta.only and field.name in self._meta.exclude:
                continue

            # 等价于 model_obj.get(field.name)
            field_value = getattr(model_obj, field.name)  # 获取字段的值
            process_func = getattr(self, field.name, Field.process)  # 获取（在描述符中）字段的处理方法
            if field.choices:  # 获取该字段的display
                data_dict[field.name + '_display'] = dict(field.choices).get(field_value, None)
            data_dict[field.name] = process_func(field_value)

        # model._meta.fields以外的字段，多对多，关联表，和注解等无法通过遍历模型的_meta.fields获取
        for field_name in self._out_fields:
            if self._meta.only and field_name not in self._meta.only:
                continue
            if not self._meta.only and field_name in self._meta.exclude:
                continue

            process = getattr(self, field_name)
            data_dict[field_name] = process(model_obj)

        return data_dict



from django.db.models import QuerySet


class PaginatorMeta(type):
    def __new__(mcs, name, bases, act):
        mcs.const(name, act)
        return super().__new__(mcs, name, bases, act)

    @staticmethod
    def const(name, act):
        if 'page_param' in act and not isinstance(act['page_param'], str):
            raise TypeError("{}的page_param必须是str类型！".format(name))

        if 'page_size_param' in act and not isinstance(act['page_size_param'], str):
            raise TypeError("{}的page_size_param必须是str类型！".format(name))

        if 'max_page_size' in act and not isinstance(act['max_page_size'], int):
            raise TypeError("{}的max_page_size必须是int类型！".format(name))


class PaginatorBase(metaclass=PaginatorMeta):
    """适用Django翻页功能"""
    max_page_size = 100

    def __init__(self, queryset, page, page_size=None):
        """
        :param queryset:QuerySet
        :param page: 以为整型，也可以是被转为整型的字符串，为None则默认1
        :param page_size: 以为整型，也可以是被转为整型的字符串，为None则默认30
        """
        if not isinstance(queryset, QuerySet):
            raise TypeError("queryset必须是QuerySet对象")
        self._queryset = queryset
        self._total = queryset.count()  # 数据总条数
        # 每页获取条数
        self._page_size = self._process_page_size(page_size)
        # 取整数页，先忽略可能的余数
        page_count = self._total // self._page_size
        # 如果总数能被page_size整数意味着没有余数，否则总页数+1
        self._page_count = page_count if self._total % self._page_size == 0 else page_count + 1  # 总页数
        # _page_count计算完成才能计算page page不能大于_page_count，也不能小于1
        self._page = self._process_page(page)

    def _process_page(self, page):
        """将传进来的page处理成合法值"""
        page = page if page else 1  # 如果传进来None值
        page = self._to_int(page)
        page = min(self._page_count, page)  # 页码不能大于总页数
        page = max(1, page)  # 防止page_count（空数据）为0时，page也为0
        return page

    def _process_page_size(self, page_size):
        page_size = self._to_int(page_size) if page_size else 30
        if self.max_page_size:  # 如果设置了max_size,则不能大于max_size
            page_size = min(page_size, self.max_page_size)
        page_size = min(page_size, self._total)  # page_size不能大于总条数
        # 字符串'0'会引发除0错误
        page_size = max(1, page_size)  # page_size不能小于1
        return page_size

    def sub_set(self):
        page = self._page
        start_index, end_index = (page - 1) * self._page_size, page * self._page_size
        sub_set = self._queryset[start_index:end_index]
        return sub_set

    def _to_int(self, value):
        try:
            return int(value)  # 如果传进来字符串值
        except Exception as err:
            raise ValueError('将{} 转为整型时失败！'.format(str(value)))

    def page_count(self):
        return {'page': self._page,
                'page_size': self._page_size,
                'page_count': self._page_count,
                'total': self._total,
                'pre': None if self._page == 1 else self._page - 1,
                'next': self._page + 1 if self._page < self._page_count else None
                }



from .paginator_base import PaginatorBase


class Paginator(PaginatorBase):
    page_param = 'page'  # 查询参数名
    page_size_param = 'page_size'
    max_page_size = 100

    def __init__(self, queryset, request):
        query = getattr(request, request.method)
        page = query.get(self.page_param)
        page_size = query.get(self.page_size_param)
        super().__init__(queryset, page, page_size)



from .fields import Field
from django.db.models import QuerySet, Model

__all__ = ['Filter']


class FilterMeta(type):
    def __new__(mcs, name, bases, attrs):
        # 所有metaclass=FilterMeta的类，都是FilterMeta的实例（元类的实例是类）
        parents = [b for b in bases if isinstance(b, FilterMeta)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        _meta = mcs.build_meta(name, parents, attrs)
        # meta设置默认值
        _meta = mcs.meta_set_defaults(_meta)
        # 使用描述符定义的字段元组
        _meta.desc_fields = mcs.build_desc_fields(name, parents, attrs)
        # 处理meta中的一些值
        _meta = mcs.process_meta(_meta)
        new_class = super().__new__(mcs, name, bases, attrs)
        new_class._meta = _meta
        return new_class

    @classmethod
    def build_meta(mcs, name, parents: list, attrs: dict):
        """
        生成类的_meta信息
        :param name:类名称
        :param parents:继承的父类列表
        :param attrs:属性字典
        :return:meta class
        """
        own_meta = attrs.pop('Meta', None)  # 自身定义的meta class
        parent_meta = getattr(parents[-1], '_meta', None)  # 父类的meta class
        if own_meta:
            # 如果自身没有在Meta中定义model，且父类也没有定义，会抛出异常
            if not getattr(own_meta, 'model', None) and not getattr(parent_meta, 'model', None):
                raise AttributeError(f'过滤类:{name} 没有在Class Meta中指定model')
            elif not getattr(own_meta, 'model', None):  # 如果自身没有在Meta中定义model则从父类中继承（右侧继承）
                own_meta.model = parent_meta.model
        meta = own_meta or parent_meta
        # 深度复制meta, 防止子类继承父类的meta时，子类对meta进行修改影响到父类
        meta = type('_meta', (), dict(meta.__dict__))
        return meta

    @classmethod
    def meta_set_defaults(mcs, meta):
        """meta设置默认值"""
        meta_defaults = dict(fields=(),
                             contains_fields=(),
                             ordering_fields=(),
                             default_ordering=None,
                             ordering_param='ordering',
                             manager=None)
        # 当子组件没有设置meta的部分属性时给一个默认值
        for attr_name, value in meta_defaults.items():
            if not getattr(meta, attr_name, None):
                setattr(meta, attr_name, value)
        return meta

    @classmethod
    def process_meta(mcs, meta):
        """处理meta中的一些值"""
        # 处理fields ，__all__代表支持全部字段支持过滤
        if meta.fields == '__all__':
            meta.fields = (field.name for field in meta.model._meta.fields)
        # 排除以描述符方式定义的查询字段
        meta.fields = tuple(i for i in meta.fields if i not in meta.desc_fields)

        # 处理ordering_fields ,__all__代表支持全部字段支持排序
        if meta.ordering_fields == '__all__':
            meta.ordering_fields = tuple(field.name for field in meta.model._meta.fields)

        # 如果没有设置管理器取基础管理器
        if not meta.manager:
            meta.manager = meta.model._base_manager
        elif isinstance(meta.manager, str):
            try:
                meta.manager = getattr(meta.model, meta.manager)
            except AttributeError:
                raise AttributeError(f'模型管理器{meta.manager}不存在')
        return meta

    @classmethod
    def build_desc_fields(mcs, name, parents, attrs) -> tuple:
        """获取类中以描述符方式定义的查询字段的名称元组"""
        desc_fields = set()
        for attr_name, attr in attrs.items():
            if isinstance(attr, Field):
                desc_fields.add(attr_name)
        # 父类中的desc_fields
        for p in parents:
            parent_desc_fields = getattr(getattr(p, '_meta', None), 'desc_fields', ())
            desc_fields = set(i for i in parent_desc_fields) | desc_fields
        return tuple(desc_fields)

    def __call__(cls, *args, **kwargs):
        self = super().__call__(*args, **kwargs)
        return self.result_queryset


class Filter(metaclass=FilterMeta):
    ORDERING_PARAM = 'ordering'

    def __init__(self, user_query, queryset=None):
        if queryset is None:
            self.result_queryset = self._meta.manager.all()
        elif isinstance(queryset, QuerySet):
            self.result_queryset = queryset
        else:
            raise ValueError(f'{self.__class__.__name__} queryset必须是QuerySet Object')
        # 生成用来查询的条件
        query = self._generate_query(user_query)
        # 进行过滤处理
        self.result_queryset = self.result_queryset.filter(**query)
        # 进行排序处理
        order_by = user_query.get(self.ORDERING_PARAM)
        self.result_queryset = self._process_ordering(order_by, self.result_queryset)

    def _generate_query(self, user_query) -> dict:
        """生成查询条件"""
        query = {}
        # 收集用户查询条件，这一步收集meta.fields中定义的field
        for key, value in user_query.items():
            if key in self._meta.fields:
                key = key + '__icontains' if key in self._meta.contains_fields else key
                query[key] = value

        # 收集用户查询条件，这一步收集描述符中的查询条件
        for key, value in user_query.items():
            if key in self._meta.desc_fields:  # 描述符中定义的字段名称列表
                setattr(self, key, value)  # 赋值描述符
                query.update(getattr(self, key))  # 获取描述符得到一个字典
        return query

    def _process_ordering(self, order_by: str or list, queryset: QuerySet) -> QuerySet:
        if not isinstance(order_by, (str, list)):
            order_by = []
        elif isinstance(order_by, str):
            order_by = order_by.split(',')  # 到这里ordering已经是是列表
        # order_by，只剩允许排序的字段
        order_by = tuple(filter(lambda i: i.strip('-') in self._meta.ordering_fields, order_by))
        if order_by:
            return queryset.order_by(*order_by)
        else:
            return queryset



from distutils.core import setup

# from os import path as os_path
#
# this_directory = os_path.abspath(os_path.dirname(__file__))
#
#
# def read_file(filename):
#     with open(os_path.join(this_directory, filename)) as f:
#         long_description = f.read()
#         return long_description
from setuptools import Extension, dist, find_packages, setup

setup(
    name="django-fate",
    version="0.30",
    author="Ryan",
    packages=find_packages(),
    author_email="604729765@qq.com",
    url="https://www.leetab.com",
    description="Django项目中使用的简易API模块",
    # install_requires=["redis>=3.5.3"],
    # long_description=open("README.rst", encoding="utf-8").read(),
)

# python setup.py sdist
# twine upload dist/*



from django import forms

__all__ = ['pick_form']


def pick_form(*fields: str, target_form) -> forms.Form:
    """根据fields提供的字段名，动态地从target_form中摘取所需字段，生成并返回一个Form Class
    :param fields: 生成Form的所需字段名
    :param target_form:摘取的目标Form
    :return:Form Class

    class AppForm(forms.Form):
        name = forms.CharField(required=True)
        phone = forms.CharField(required=True, min_length=11)
        username = forms.CharField(required=True)
        password = forms.CharField(required=True)
        .....

    LoginForm =pick_form('username','password',target_form=AppForm)
    等价于:
     class LoginForm(forms.Form):
        username = forms.CharField(required=True)
        password = forms.CharField(required=True)

    """
    if not issubclass(target_form, forms.Form):
        raise TypeError(f'{target_form.__name__}必须是Form的子类！')

    # 目标Form类中定义的所有字段字典
    target_fields = target_form.base_fields
    form_fields = {}  # 从目标类中收集的字段字典
    for field_name in fields:
        try:
            form_fields[field_name] = target_fields[field_name]
        except KeyError:
            raise ValueError(f'pick_form接受了一个{field_name}参数,但是{target_form.__name__}中未定义{field_name}')
        # clean部分
        try:
            form_fields['clean_' + field_name] = getattr(target_form, 'clean_' + field_name)
        except AttributeError:
            pass

    return type('Form', (forms.Form,), form_fields)
